#!/usr/bin/perl -w

#
# Copyright (C) 2006 by Victor Julien <victor@inliniac.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

# Drop-in replacement for Barnyard for feeding ModSecurity
# alerts to the Sguil NSM system.
# 
# Reads files or symlinks from the queue directory, feeds them
# to Sguil and removes the file/symlink.
#

use strict;
use warnings;
use diagnostics;

use IO::Socket;
use DirHandle;
use Getopt::Std;
use Time::Local;

use strict "vars";
use strict "subs";

my $version = "0.5";
my $license = "GNU GPL. See http://www.gnu.org/licenses/gpl.txt for more information";

#
# CONVERSION FUNCTIONS
#

sub ConvertASCIItoHEX
{
	my $asciistr = shift @_;
	
	#(my $str = shift) =~ s/(.|\n)/sprintf("%02lx", ord $asciistr)/eg;
	my $hexstr = join '', unpack "H*", $asciistr;

	#print $hexstr . "\n";

	return($hexstr);
}

sub ConvertSEVERITYtoPRIO
{
	my $severity = shift @_;
	
	# modsec has severity:	emergency, alert, critical, error,
	# 			warning, notice, info, debug
	if($severity eq "EMERGENCY") {
		return 7;
	} elsif ($severity eq "ALERT") {
		return 6;
	} elsif ($severity eq "CRITICAL") {
		return 5;
	} elsif ($severity eq "ERROR") {
		return 4;
	} elsif ($severity eq "WARNING") {
		return 3;
	} elsif ($severity eq "NOTICE") {
		return 2;
	} elsif ($severity eq "INFO") {
		return 1;
	} else {
		return 0;
	}
}

sub ConvertMONTHSTRtoDEC
{
	my $monthstr = shift;

	# stupid trick to convert month(str) to month(dec)
	if( $monthstr eq "Jan" ) {	return 1;
	} elsif( $monthstr eq "Feb" ) {	return 2;
	} elsif( $monthstr eq "Mar" ) {	return 3;
	} elsif( $monthstr eq "Apr" ) {	return 4;
	} elsif( $monthstr eq "May" ) {	return 5;
	} elsif( $monthstr eq "Jun" ) {	return 6;
	} elsif( $monthstr eq "Jul" ) {	return 7;
	} elsif( $monthstr eq "Aug" ) {	return 8;
	} elsif( $monthstr eq "Sep" ) {	return 9;
	} elsif( $monthstr eq "Oct" ) {	return 10;
	} elsif( $monthstr eq "Nov" ) {	return 11;
	} elsif( $monthstr eq "Dec" ) {	return 12; }

	return -1;
}

sub ConvertIPtoDEC
{
	my $ip = shift;

	( my $one , my $two , my $three , my $four ) = split(/\./, $ip);
	my $dec = $one * 16777216 + $two * 65536 + $three * 256 + $four;
	
	return($dec);
}

#
# COMMUNICATION FUNCTIONS
#


sub SendSidCidRequest
{
	my $my_sock = shift @_;
	my $my_sensor_name = shift @_;

	$my_sock->send("SidCidRequest $my_sensor_name\n");
}

sub RecvSidCidResponse
{
	my $my_sock = shift @_;
	my $recv_buf = readline( $my_sock );
	my %array;

	print "recv_buf " . $recv_buf . "\n";

	(my $cmd, $array{"sid"}, $array{"cid"} ) = split (/ /, $recv_buf);

	#print "cmd " . $cmd . " sid " . $array{"sid"} . " cid " . $array{"cid"} . "\n";

	return(%array);
}

#
# Call with: sock, sid, cid, sensorname, alert hash
#
# Returns:	-1: on error
# 		0: on success
#
sub SendRTEVENT
{
	my $my_sock = shift @_;
	my $sid = shift @_;
	my $cid = shift @_;
	my $sensor_name = shift @_;

	# get the alert hash from the caller (by reference)
	my $ref = shift @_;
	my %alert = %{$ref};

	my $status = 0;
	my $snort_event_id = 0;
	my $snort_event_ref = 0;

	# time.
	my $modsec_time = $alert{"year"} . "-" . $alert{"month"} . "-" . $alert{"day"} . " " . $alert{"timestr"};
	my $modsec_ts = $modsec_time;
	
	my $modsec_siggen = 0;
	my $modsec_sigid = 0;

	my $modsec_rev = $alert{"rev"};
	my $modsec_msg = $alert{"msg"};
	my $modsec_prio = $alert{"prio"};
	my $modsec_class  = $alert{"class"};
	# IPaddresses
	my $modsec_sip_dec = $alert{"sipdec"};
	my $modsec_sip_str = $alert{"sipstr"};
	my $modsec_dip_dec = $alert{"dipdec"};
	my $modsec_dip_str = $alert{"dipstr"};
	# IP header
	my $modsec_ipver = 4;
	my $modsec_ipproto = 6;	
	# rest of the IP header -- we will never get this from ModSecurity
	my $modsec_ip = "0 0 0 0 0 0 0 0";
	# ICMP -- we will never get this from ModSecurity
	my $modsec_icmp = "0 0 0 0 0";
	# ports
	my $modsec_sp = $alert{"sp"};
	my $modsec_dp = $alert{"dp"};
	# TCP/UDP header -- we will never get this from ModSecurity
	my $modsec_tcp = "0 0 0 0 0 0 0 0";
	my $modsec_udp = "0 0";
	# convert the entire alert file to hex and append it as the the payload
	my $modsec_payload = $alert{"payload"};

	#print $modsec_payload . "\n";

	# assemble the string to send to the sensor.
	# 
	#           0       1       2    3    4            5               6
	my $str = "RTEVENT $status $sid $cid $sensor_name $snort_event_id $snort_event_ref " . 
	#  7              8              9             10            11
	"\{$modsec_time\} $modsec_siggen $modsec_sigid $modsec_rev \{$modsec_msg\} " .
	#  12            13           14            15              16
	"\{$modsec_ts\} $modsec_prio $modsec_class $modsec_sip_dec $modsec_sip_str " .
	# 17               18              19              20            21-28
	"$modsec_dip_dec $modsec_dip_str $modsec_ipproto $modsec_ipver $modsec_ip " .
	# 29-33         34         35         36-43       44-45        46
	"$modsec_icmp $modsec_sp $modsec_dp $modsec_tcp $modsec_udp \{$modsec_payload\}\n";

	$my_sock->send($str);
}

sub RecvRTEVENTResponse
{
	my $my_sock = shift @_;
	my $recv_buf = readline( $my_sock );

	print "recv_buf " . $recv_buf . "\n";

	(my $cmd, my $cid ) = split (/ /, $recv_buf);

	#print "cmd " . $cmd . " sid " . $array{"sid"} . " cid " . $array{"cid"} . "\n";

	return($cid);
}

#
# ALERT PROCESSING
#

#
# Some preprocessing before sending the alert
# to Sguil.
#
# This function is only called for actual alerts.
#
# Returns:	-1 on error
# 		0 on success
#
sub PreprocessAlert
{
	# get the alert hash from the caller (by reference)
	my $ref = shift @_;

	#
	# SECTION A
	#
	
	# stupid trick to convert month(str) to month(dec)
	my $month = ConvertMONTHSTRtoDEC $ref->{"monthstr"};
	if($month != -1) {
		$ref->{"month"} = $month;
	} else {
		print "Month parse failed.\n";
		return -1;
	}

	# convert sipstr and dipstr to dec as well
	$ref->{"sipdec"} = ConvertIPtoDEC $ref->{"sipstr"};
	$ref->{"dipdec"} = ConvertIPtoDEC $ref->{"dipstr"};

	# timezone handling: the time/date is logged like this:
	# 13/Aug/2006:21:59:24 +0200, which is 19:59.24 UTC. In
	# Sguil we want the UTC value, so we convert it here.
	# 
	# first parse the timestring
	(my $hour, my $min, my $sec ) = split(/:/, $ref->{"timestr"});

	# get the unixtime in gm, month is from 0 to 11.
	my $time = timegm($sec,$min,$hour,$ref->{"day"},$ref->{"month"}-1,$ref->{"year"});
					
	# parse the timezone offset. It looks like +0200 or -0100
	$_ = $ref->{"tz"};
	my @parse = /(.{1})(\d{2})(\d{2})$/;
	if(@parse == 0) {
		print "Parsing timezone information failed.\n";
		return -1;
	} else {
		(my $op, my $tz_hour, my $tz_min) = @parse;
		my $tz_val = $tz_hour * 3600 + $tz_min * 60;

		# +0200 means we have to subtrackt 2 hrs to get UTC
		if ( $op eq "+" ) {
			$time -= $tz_val;
		} elsif ( $op eq "-" ) {
			$time += $tz_val;
		} else {
			print "Unknown operator $op in timezone string " . $ref->{"tz"} . "\n";
			return -1;
		}

		#print "time after tz apply: $time, gmtime " . gmtime($time) . "\n";

		# convert the new unix time
		# Fri Aug 18 04:35:35 2006
		# Mon Aug  7 07:51:23 2006 <= note two spaces, split(/ / won't work
		# that is what the \ ? takes care of.
		# 
		# update day, timestr and year
		$_ = gmtime($time);
		@parse = /(\S*) (\S*) \ ?(\d*) (.*) (\d+)/;
		if ( @parse == 0 ) {
			print "Date/time parsing error, parsing $_ failed\n";
			return -1;
		} else {
			(my $dayname, $ref->{"monthstr"}, $ref->{"day"}, $ref->{"timestr"}, $ref->{"year"} ) = @parse;
		}

		# update month
		$month = ConvertMONTHSTRtoDEC $ref->{"monthstr"};
		if($month != -1) {
			$ref->{"month"} = $month;
		} else {
			print "Month parse failed, parsing " . $ref->{"monthstr"} . ".\n";
			return -1;
		}

		#print $ref->{"timestr"} . "\n";
	}

	# add a padding zero to match the by behaviour.
	if ( $ref->{"day"} < 10 ) {
		$ref->{"day"} = "0" . $ref->{"day"};
	}
	if ( $ref->{"month"} < 10 ) {
		$ref->{"month"} = "0" . $ref->{"month"};
	}

	#
	# SECTION H
	#
	
	# check for a message with a message, get it?
	if( defined( $ref->{"themsg"} ) ) {
		#print "themsg: " .  $ref->{"themsg"} . "\n";

		$_ = $ref->{"themsg"};

		# loop through all options which have the format:
		# Pattern match "!^HTTP/(0\\.9|1\\.0|1\\.1)$" at SERVER_PROTOCOL [id "340000"] [rev "1"] [msg "Bad HTTP Protocol"]
		# as long as we have options, $nextparm and $nextval will be set to:
		# id and 340000, rev and 1, etc
		while ( m/.* \[.*\].*/ )
		{
			#print "match\n";
			my @parsed_str = /.*?\[(.*?) \"(.*?)\"\](.*)/;
			if( @parsed_str != 0 )
			{
				( my $nextparm, my $nextval, my $therest ) = @parsed_str;
				#print "nextparm " . $nextparm . "\nnextval " . $nextval . "\nthe rest " . $therest . "\n";

				# update $_ for the regex in our while loop
				$_ = $therest;
	
				# lets see what we have
				if ($nextparm eq "msg") {
					#print "we have a msg: $nextval\n";
					$ref->{"themsg"} = $nextval;
				} elsif ($nextparm eq "id") {
					#print "we have an id: $nextval\n";
					$ref->{"id"} = $nextval;
				} elsif ($nextparm eq "rev") {
					#print "we have a rev: $nextval\n";
					$ref->{"rev"} = $nextval;
				} else {
					print "Warning: unknown keyword $nextparm with value $nextval in " . $ref->{"themsg"} . ".\n";
				}
			} else {
				# this will break us out of the while loop
				$_ = "";
			}
		}
	}

	# severity relates to priority.
	$ref->{"prio"} = ConvertSEVERITYtoPRIO $ref->{"severity"};

	# lets see what kind of class we are going to use
	# we use *-attack for stuff that is blocked, *-activity
	# for the rest
	if($ref->{"code"} < 400) {
		$ref->{"class"} = "web-application-activity";
	} else {
		$ref->{"class"} = "web-application-attack";
	}
	
	# Event message
	$ref->{"msg"} = "MOD-SECURITY " . $ref->{"code"} . " " . $ref->{"themsg"};

	# hex the payload
	$ref->{"payload"} = ConvertASCIItoHEX $ref->{"file"};
	
	return 0;
}

#
# Returns	-1: error
# 		0 : success
#
sub ParseAlertFile
{
	my $my_filename = shift @_;
	my $ref = shift @_;	# reference to event_hash

	my $line_num = 1;
	my $fileid;
	my $section = "-";
	my %event;

	# initialize some vars in case the alert doesn't contain them
	$ref->{"file"} = "";
	$ref->{"severity"} = "DEBUG"; # lowest by modsec
	$ref->{"rev"} = 0;
	
	open (FILE , $my_filename) or die "Could not open file $my_filename: $!\n";

	while(<FILE>)
	{
		# store all lines of the file so we can use it
		# as the alert payload later
		$ref->{"file"} = $ref->{"file"} . $_;

		# Get this file's unique id
		if($line_num == 1)
		{
			my @parsed_line = /^--(\S+)-A--$/;
			if ( @parsed_line == 0 ) {
				goto error;
			} else {
				#print "parsed line @parsed_line\n";
				
				( $fileid ) = @parsed_line;
				#print "fileid " . $fileid . "\n";
			}
		}

		# parse the section character
		if ( /^--$fileid-\S+--$/ ) {
			my @parsed_line = /^--$fileid-(\S+)--$/;
			( $section ) = @parsed_line;
			#print "section " . $section . "\n";
		}
		
		#section A -- time, ipaddresses and ports
		if($section eq "A")
		{
			if(m/^\[.*\] .*$/) {
				my @parsed_line = /^\[(\d+)\/(\S+)\/(\d+):(.*) (.*)\] (\S+) (\S+) (\d+) (\S+) (\d+)$/;
				if ( @parsed_line == 0 ) {
					goto error;
				} else {
					( $ref->{"day"}, $ref->{"monthstr"}, $ref->{"year"},
					  $ref->{"timestr"}, $ref->{"tz"}, $ref->{"uniqueid"}, 
					  $ref->{"sipstr"}, $ref->{"sp"}, $ref->{"dipstr"},
					  $ref->{"dp"} ) = @parsed_line;

					# all further processing is done only if we know this
					# is an alert. PreprocessAlert processes it further.
				}
			}
		}
		
		#section H, alert info, mod_sec response
		if ( $section eq "H" )
		{
			if ( m/^Message\: Access denied with code.*/ )
			{
				my @parsed_line = /^Message\: Access denied with code (\d+)\. (.*) \[severity \"(\S+)\"\]$/;
				if ( @parsed_line == 0 )
				{
					goto error;
				}
				else
				{
					( $ref->{"code"} , $ref->{"themsg"} , $ref->{"severity"} ) = @parsed_line;
				}
			}
			elsif ( m/^Message\: Access denied with redirect to.*/ )
			{
				my @parsed_line = /^Message\: Access denied with redirect to \[(.*)\]\. (.*) \[severity \"(\S+)\"\]$/;
				if ( @parsed_line == 0 )
				{
					goto error;
				}
				else
				{
					( $ref->{"redirecturl"} , $ref->{"themsg"} , $ref->{"severity"} ) = @parsed_line;
				}
			}

			# if we do not yet have a code (e.g. in a redirect case)
			# we try to find it here
			if(not defined($event{"code"}) )
			{
				if(m/^Action\: Intercepted \((\d+)\).*/)
				{
					my @parsed_line = /^Action\: Intercepted \((\d+)\)$/;
					if ( @parsed_line == 0 )
					{
						goto error;
					}
					else
					{
						( my $code ) = @parsed_line;
						#print "Action " . $code . "\n";
						$ref->{"code"} = $code;
					}
				}
			}
		}

		$line_num++;
	}

	close(FILE);
	return 0;

	# a bit ugly, but it saves us from duplicating a lot of code:
	# close file, chomp, print error, return -1.
error:
	close(FILE);
	chomp();
	print "Error parsing line $line_num: $_\n";
	return -1;
}

#
# Returns a list of filenames of event files.
# Returns 0 if no files were found.
#
sub GetAlertFiles {
	my $dir = shift;

	#print $dir . "\n";
	
	my $dh = DirHandle->new($dir) or die "can't open dir $dir: $!\n";
	return  sort
		grep { -f        }
		grep { /^$dir\/modsec\.log\..*/ } # we want only the files that start with modsec.log.
		map  { "$dir/$_" }
		grep { !/^\./    }
		$dh->read();
}

#
# START MAIN
#

#
# option parsing
# 
my %option = ();
getopts("n:l:r:tv", \%option);

if ( defined $option{v} ) {
	print "\n";
	print "modsec_by.pl version $version.\n\n";
	print "Copyright (c) 2006 by Victor Julien <victor\@inliniac.net>.\n";
	print "Release under $license.\n";
	print "\n";
	exit 0;
}

$option{n} = ""    unless defined $option{n}; # sensor name
$option{l} = ""    unless defined $option{l}; # logdir
$option{r} = "yes" unless defined $option{r}; # remove the processed files?

my %sensor = ();
$sensor{"name"}   = $option{n};
$sensor{"dir"}    = $option{l};
$sensor{"remove"} = $option{r};
if(defined($option{t})){
	$sensor{"testfiles"} = "true";
} else {
	$sensor{"testfiles"} = "false";
}


if ( ( $sensor{"name"} eq "" ) || ( $sensor{"dir"} eq "" ) ) {
	print "Please supply:\n\tthe sensor name (-n <sensor name)\n\tthe log directory (-l <directory>)\n";
	exit(0);
}

# connect only if we are not in test mode
if ( $sensor{"testfiles"} eq "false" )
{
	#
	# Connect and request the SID and CID
	#
	$sensor{"sock"} = new IO::Socket::INET (
			PeerAddr => '127.0.0.1',
			PeerPort => '7735',
			Proto => 'tcp',
			) or die "Could not connect: $!\n";

	#
	# Get the SID and CID.
	#
	SendSidCidRequest ( $sensor{"sock"} , $sensor{"name"} );
	my %result = RecvSidCidResponse ( $sensor{"sock"} );
	%sensor = (%sensor , %result);
	print "!!! " . $sensor{"sid"} . " " . $sensor{"cid"} . "\n";
}


#
# Main loop
#
#
while ( 1 )
{
	my @list = GetAlertFiles $sensor{"dir"};
	if ( @list == 0 ){
		print "listsize is 0, sleeping 3 seconds.\n";
		sleep 3;
	} else {
		# parse each file and check if it really was an alert
		foreach my $file (@list) {
			print "file is $file\n";

			# event hash for all data for the alert, called
			# event hash because we are not yet sure if this
			# if an alert.
			my %event_hash = ();
			# create a reference to work with
			my $hash_ref = \%event_hash;

			if ( ParseAlertFile( $file , $hash_ref ) == 0 )
			{
				# if we have no code, it was no alert
				if(exists($event_hash{"code"}))
				{
					print $event_hash{"code"} . "\n";

					if ( PreprocessAlert ( $hash_ref ) == 0 ) {
						print $event_hash{"themsg"} . "\n";
			
						if( $sensor{"testfiles"} eq "false" )
						{
							# increase the cid so the alert gets it's uniq cid.
							$sensor{"cid"}++;

							# Send the alert to Sguil
							# TODO: error checking
							SendRTEVENT ( $sensor{"sock"} , $sensor{"sid"} , $sensor{"cid"} , $sensor{"name"}, $hash_ref );
							# get the response
							# TODO: error checking
							my $ret_cid = RecvRTEVENTResponse ( $sensor{"sock"} );
						}
					}
				}
			}

			# remove only if we want to (useful for debugging)
			if( $sensor{"remove"} eq "yes" && $sensor{"testfiles"} eq "false" ) {
				unlink $file;
			}
		}
	}

	# in test mode we just look at all files once
	if ( $sensor{"testfiles"} eq "true" ) {
		exit 0;
	}
}

